
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "config.h"

#include <gdk-pixbuf/gdk-pixbuf.h>
/* enumerations from "gdk-pixbuf-core.h" */
GType
gdk_pixbuf_alpha_mode_get_type (void)
{
    static gsize g_define_type;

    if (g_once_init_enter (&g_define_type)) {
        static const GEnumValue values[] = {
            { GDK_PIXBUF_ALPHA_BILEVEL, "GDK_PIXBUF_ALPHA_BILEVEL", "bilevel" },
            { GDK_PIXBUF_ALPHA_FULL, "GDK_PIXBUF_ALPHA_FULL", "full" },
            { 0, NULL, NULL }
        };
        GType enum_type = g_enum_register_static (g_intern_static_string ("GdkPixbufAlphaMode"), values);

        g_once_init_leave (&g_define_type, enum_type);
    }
    return g_define_type;
}
GType
gdk_colorspace_get_type (void)
{
    static gsize g_define_type;

    if (g_once_init_enter (&g_define_type)) {
        static const GEnumValue values[] = {
            { GDK_COLORSPACE_RGB, "GDK_COLORSPACE_RGB", "rgb" },
            { 0, NULL, NULL }
        };
        GType enum_type = g_enum_register_static (g_intern_static_string ("GdkColorspace"), values);

        g_once_init_leave (&g_define_type, enum_type);
    }
    return g_define_type;
}
GType
gdk_pixbuf_error_get_type (void)
{
    static gsize g_define_type;

    if (g_once_init_enter (&g_define_type)) {
        static const GEnumValue values[] = {
            { GDK_PIXBUF_ERROR_CORRUPT_IMAGE, "GDK_PIXBUF_ERROR_CORRUPT_IMAGE", "corrupt-image" },
            { GDK_PIXBUF_ERROR_INSUFFICIENT_MEMORY, "GDK_PIXBUF_ERROR_INSUFFICIENT_MEMORY", "insufficient-memory" },
            { GDK_PIXBUF_ERROR_BAD_OPTION, "GDK_PIXBUF_ERROR_BAD_OPTION", "bad-option" },
            { GDK_PIXBUF_ERROR_UNKNOWN_TYPE, "GDK_PIXBUF_ERROR_UNKNOWN_TYPE", "unknown-type" },
            { GDK_PIXBUF_ERROR_UNSUPPORTED_OPERATION, "GDK_PIXBUF_ERROR_UNSUPPORTED_OPERATION", "unsupported-operation" },
            { GDK_PIXBUF_ERROR_FAILED, "GDK_PIXBUF_ERROR_FAILED", "failed" },
            { GDK_PIXBUF_ERROR_INCOMPLETE_ANIMATION, "GDK_PIXBUF_ERROR_INCOMPLETE_ANIMATION", "incomplete-animation" },
            { 0, NULL, NULL }
        };
        GType enum_type = g_enum_register_static (g_intern_static_string ("GdkPixbufError"), values);

        g_once_init_leave (&g_define_type, enum_type);
    }
    return g_define_type;
}
/* enumerations from "gdk-pixbuf-transform.h" */
GType
gdk_interp_type_get_type (void)
{
    static gsize g_define_type;

    if (g_once_init_enter (&g_define_type)) {
        static const GEnumValue values[] = {
            { GDK_INTERP_NEAREST, "GDK_INTERP_NEAREST", "nearest" },
            { GDK_INTERP_TILES, "GDK_INTERP_TILES", "tiles" },
            { GDK_INTERP_BILINEAR, "GDK_INTERP_BILINEAR", "bilinear" },
            { GDK_INTERP_HYPER, "GDK_INTERP_HYPER", "hyper" },
            { 0, NULL, NULL }
        };
        GType enum_type = g_enum_register_static (g_intern_static_string ("GdkInterpType"), values);

        g_once_init_leave (&g_define_type, enum_type);
    }
    return g_define_type;
}
GType
gdk_pixbuf_rotation_get_type (void)
{
    static gsize g_define_type;

    if (g_once_init_enter (&g_define_type)) {
        static const GEnumValue values[] = {
            { GDK_PIXBUF_ROTATE_NONE, "GDK_PIXBUF_ROTATE_NONE", "none" },
            { GDK_PIXBUF_ROTATE_COUNTERCLOCKWISE, "GDK_PIXBUF_ROTATE_COUNTERCLOCKWISE", "counterclockwise" },
            { GDK_PIXBUF_ROTATE_UPSIDEDOWN, "GDK_PIXBUF_ROTATE_UPSIDEDOWN", "upsidedown" },
            { GDK_PIXBUF_ROTATE_CLOCKWISE, "GDK_PIXBUF_ROTATE_CLOCKWISE", "clockwise" },
            { 0, NULL, NULL }
        };
        GType enum_type = g_enum_register_static (g_intern_static_string ("GdkPixbufRotation"), values);

        g_once_init_leave (&g_define_type, enum_type);
    }
    return g_define_type;
}

/* Generated data ends here */

